/*
    FreeRTOS V7.4.0 - Copyright (C) 2013 Real Time Engineers Ltd.

    FEATURES AND PORTS ARE ADDED TO FREERTOS ALL THE TIME.  PLEASE VISIT
    http://www.FreeRTOS.org TO ENSURE YOU ARE USING THE LATEST VERSION.

    ***************************************************************************
     *                                                                       *
     *    FreeRTOS tutorial books are available in pdf and paperback.        *
     *    Complete, revised, and edited pdf reference manuals are also       *
     *    available.                                                         *
     *                                                                       *
     *    Purchasing FreeRTOS documentation will not only help you, by       *
     *    ensuring you get running as quickly as possible and with an        *
     *    in-depth knowledge of how to use FreeRTOS, it will also help       *
     *    the FreeRTOS project to continue with its mission of providing     *
     *    professional grade, cross platform, de facto standard solutions    *
     *    for microcontrollers - completely free of charge!                  *
     *                                                                       *
     *    >>> See http://www.FreeRTOS.org/Documentation for details. <<<     *
     *                                                                       *
     *    Thank you for using FreeRTOS, and thank you for your support!      *
     *                                                                       *
    ***************************************************************************


    This file is part of the FreeRTOS distribution.

    FreeRTOS is free software; you can redistribute it and/or modify it under
    the terms of the GNU General Public License (version 2) as published by the
    Free Software Foundation AND MODIFIED BY the FreeRTOS exception.

    >>>>>>NOTE<<<<<< The modification to the GPL is included to allow you to
    distribute a combined work that includes FreeRTOS without being obliged to
    provide the source code for proprietary components outside of the FreeRTOS
    kernel.

    FreeRTOS is distributed in the hope that it will be useful, but WITHOUT ANY
    WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
    FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
    details. You should have received a copy of the GNU General Public License
    and the FreeRTOS license exception along with FreeRTOS; if not itcan be
    viewed here: http://www.freertos.org/a00114.html and also obtained by
    writing to Real Time Engineers Ltd., contact details for whom are available
    on the FreeRTOS WEB site.

    1 tab == 4 spaces!

    ***************************************************************************
     *                                                                       *
     *    Having a problem?  Start by reading the FAQ "My application does   *
     *    not run, what could be wrong?"                                     *
     *                                                                       *
     *    http://www.FreeRTOS.org/FAQHelp.html                               *
     *                                                                       *
    ***************************************************************************


    http://www.FreeRTOS.org - Documentation, books, training, latest versions,
    license and Real Time Engineers Ltd. contact details.

    http://www.FreeRTOS.org/plus - A selection of FreeRTOS ecosystem products,
    including FreeRTOS+Trace - an indispensable productivity tool, and our new
    fully thread aware and reentrant UDP/IP stack.

    http://www.OpenRTOS.com - Real Time Engineers ltd license FreeRTOS to High
    Integrity Systems, who sell the code with commercial support,
    indemnification and middleware, under the OpenRTOS brand.

    http://www.SafeRTOS.com - High Integrity Systems also provide a safety
    engineered and independently SIL3 certified version for use in safety and
    mission critical applications that require provable dependability.
*/

/******************************************************************************
 * >>>>>> NOTE 1: <<<<<<
 *
 * main() can be configured to create either a very simple LED flasher demo, or
 * a more comprehensive test/demo application.
 *
 * To create a very simple LED flasher example, set the
 * mainCREATE_SIMPLE_LED_FLASHER_DEMO_ONLY constant (defined below) to 1.  When
 * this is done, only the standard demo flash tasks are created.  The standard
 * demo flash example creates three tasks, each of which toggle an LED at a
 * fixed but different frequency.
 *
 * To create a more comprehensive test and demo application, set
 * mainCREATE_SIMPLE_LED_FLASHER_DEMO_ONLY to 0.
 *
 * >>>>>> NOTE 2: <<<<<<
 *
 * In addition to the normal set of standard demo tasks, the comprehensive test
 * makes heavy use of the floating point unit, and forces floating point
 * instructions to be used from interrupts that nest three deep.  The nesting
 * starts from the tick hook function, resulting is an abnormally long context
 * switch time.  This is done purely to stress test the FPU context switching
 * implementation, and that part of the test can be removed by setting
 * configUSE_TICK_HOOK to 0 in FreeRTOSConfig.h.
 ******************************************************************************
 *
 * main() creates all the demo application tasks and software timers, then starts
 * the scheduler.  The web documentation provides more details of the standard
 * demo application tasks, which provide no particular functionality, but do
 * provide a good example of how to use the FreeRTOS API.
 *
 */

/* A simple serial to Ethernet using LwIP and FreeRTOS. */

#include <stdio.h>

/* Kernel includes. */
#include "FreeRTOS.h"
#include "task.h"
#include "timers.h"
#include "semphr.h"

/* Demo application includes. */
#include "partest.h"
#include "flash.h"
#include "flop.h"
#include "integer.h"
#include "PollQ.h"
#include "semtest.h"
#include "dynamic.h"
#include "BlockQ.h"
#include "blocktim.h"
#include "countsem.h"
#include "GenQTest.h"
#include "QueueSet.h"
#include "recmutex.h"
#include "death.h"

/* Hardware and starter kit includes. */
#include "M480.h"

#include "lwip/netifapi.h"
#include "lwip/tcpip.h"
#include "netif/ethernetif.h"
#include "lwip/opt.h"
#include "lwip/arch.h"
#include "lwip/api.h"

#include "ste_config.h"

/* Priorities for the demo application tasks. */
#define mainFLASH_TASK_PRIORITY            ( tskIDLE_PRIORITY + 1UL )
#define mainQUEUE_POLL_PRIORITY            ( tskIDLE_PRIORITY + 1UL )
#define mainSEM_TEST_PRIORITY              ( tskIDLE_PRIORITY + 1UL )
#define mainCHECK_TASK_PRIORITY            ( tskIDLE_PRIORITY + 3UL )

#define TEST_BUILD_CLIENT    0    // 0 to build server. Only for test.
#define TEST_WITH_WINPC        1    // 1 IP set to 169.x.x.x for WinPC

//#define USE_DHCP

#if (LWIP_DHCP == 1) && defined(USE_DHCP)
#include "lwip/dhcp.h"
#endif

/*** Configure UART ports and Ethernet ***/

S_PORT_CONFIG port_config[MAX_UART_PORTS] =
{
#if MAX_UART_PORTS == 1

 #if TEST_BUILD_CLIENT == 1
  #if TEST_WITH_WINPC == 1
    {NET_CLIENT_MODE, NET_PORT_BASE+0, UART1, DEFAULT_UART_BAUD, UART_WORD_LEN_8, UART_STOP_BIT_1, UART_PARITY_NONE, 0x026CFEA9, NET_PORT_BASE+0}
  #else
    {NET_CLIENT_MODE, NET_PORT_BASE+0, UART1, DEFAULT_UART_BAUD, UART_WORD_LEN_8, UART_STOP_BIT_1, UART_PARITY_NONE, 0x0200A8C0, NET_PORT_BASE+0}
  #endif
 #else
    {NET_SERVER_MODE, NET_PORT_BASE+0, UART1, DEFAULT_UART_BAUD, UART_WORD_LEN_8, UART_STOP_BIT_1, UART_PARITY_NONE, 0x7F000001, 80}
  #endif

#elif MAX_UART_PORTS == 2
    {NET_SERVER_MODE, NET_PORT_BASE+0, UART1, DEFAULT_UART_BAUD, UART_WORD_LEN_8, UART_STOP_BIT_1, UART_PARITY_NONE, 0x7F000001, 80},
    {NET_SERVER_MODE, NET_PORT_BASE+1, UART5, DEFAULT_UART_BAUD, UART_WORD_LEN_8, UART_STOP_BIT_1, UART_PARITY_NONE, 0x7F000001, 81}

#elif MAX_UART_PORTS == 3
    {NET_SERVER_MODE, NET_PORT_BASE+0, UART1, DEFAULT_UART_BAUD, UART_WORD_LEN_8, UART_STOP_BIT_1, UART_PARITY_NONE, 0x7F000001, 80},
    {NET_SERVER_MODE, NET_PORT_BASE+1, UART5, DEFAULT_UART_BAUD, UART_WORD_LEN_8, UART_STOP_BIT_1, UART_PARITY_NONE, 0x7F000001, 81},
    {NET_SERVER_MODE, NET_PORT_BASE+2, UART0, DEFAULT_UART_BAUD, UART_WORD_LEN_8, UART_STOP_BIT_1, UART_PARITY_NONE, 0x7F000001, 82}
#endif
};

/*** Declare UART buffers ***/
uint8_t UartRxBuf[MAX_UART_PORTS][UART_RX_BUFFER_SIZE] = {0};
volatile int32_t UartRxHead[MAX_UART_PORTS] = {0};
volatile int32_t UartRxTail[MAX_UART_PORTS] = {0};
uint8_t UartTxBuf[MAX_UART_PORTS][UART_TX_BUFFER_SIZE] = {0};
volatile int32_t UartTxHead[MAX_UART_PORTS] = {0};
volatile int32_t UartTxTail[MAX_UART_PORTS] = {0};

/*** UART buffer control functions ***/

/* Return the buffer index of specified UART port */
uint32_t GetUartBufIndex(UART_T *p)
{
    uint32_t i;

    for(i=0; i<MAX_UART_PORTS; i++)
    {
        if (port_config[i].pserial == p)
            return i;
    }
    return MAX_UART_PORTS-1;
}

/* Return data length for specified head & tail of buffer */
int32_t GetBufLen(int32_t head, int32_t tail)
{
    int32_t d = head - tail;
    if (d<0) d += UART_RX_BUFFER_SIZE;
    return d;
}

/* Return data length of spcified Rx buffer */
int32_t GetRxBufLen(uint32_t index)
{
    int32_t h = UartRxHead[index];
    int32_t t = UartRxTail[index];
    return GetBufLen(h, t);
}

/* Return data length of spcified Tx buffer */
int32_t GetTxBufLen(uint32_t index)
{
    int32_t h = UartTxHead[index];
    int32_t t = UartTxTail[index];
    return GetBufLen(h, t);
}

/* Return data length of spcified UART Rx buffer */
int32_t GetUartRxBufLen(UART_T *p)
{
    return GetRxBufLen(GetUartBufIndex(p));
}

/* Return data length of spcified UART Tx buffer */
int32_t GetUartTxBufLen(UART_T *p)
{
    return GetTxBufLen(GetUartBufIndex(p));
}

/* Common UART IRQ handler */
void UART_IRQH(UART_T *pUART)
{
    uint32_t u32IntSts = pUART->INTSTS;
    uint32_t index = GetUartBufIndex(pUART);

    if (u32IntSts & (UART_INTSTS_RDAINT_Msk | UART_INTSTS_RXTOIF_Msk))
    {
        /* There are data in Rx FIFO, read them and put to circular RX buffer */

        uint8_t *pBuf = UartRxBuf[index];

        while(GetRxBufLen(index) < UART_RX_BUFFER_SIZE-1)
        {
            if (UART_IS_RX_READY(pUART))
            {
                pBuf[UartRxHead[index]] = UART_READ(pUART);
                if (++UartRxHead[index] >= UART_RX_BUFFER_SIZE)
                    UartRxHead[index] = 0;
            }
            else
                break;
        }

        while(UART_IS_RX_READY(pUART))
        {
            // Buffer is full but there are still data in FIFO
            UART_READ(pUART);
        }
    }

    if (u32IntSts & UART_INTSTS_THREINT_Msk)
    {
        uint8_t *pBuf = UartTxBuf[index];

        while(GetTxBufLen(index) && !UART_IS_TX_FULL(pUART))
        {
            UART_WRITE(pUART, pBuf[UartTxTail[index]]);
            if (++UartTxTail[index] >= UART_TX_BUFFER_SIZE)
                UartTxTail[index] = 0;
        }

        if (pUART->INTSTS & UART_INTSTS_THREINT_Msk)
        {
            UART_DISABLE_INT(pUART, UART_INTEN_THREIEN_Msk);
        }
    }
}

void UART1_IRQHandler(void)
{
    UART_IRQH(UART1);
}

void UART2_IRQHandler(void)
{
    UART_IRQH(UART2);
}

void UART3_IRQHandler(void)
{
    UART_IRQH(UART3);
}

/* Transfer data from UART to Ethernet, and vice versa. */
void exchange_data(S_PORT_CONFIG *pmap, struct netconn *conn)
{
    err_t net_err;

    struct netbuf *inbuf = NULL;
    uint8_t *n_buf;
    uint16_t n_len = 0, n_index = 0;

    uint8_t s_buf[256];
    uint16_t s_len = 0, s_index = 0;
    int32_t buf_len;
    uint32_t buf_index = GetUartBufIndex(pmap->pserial);

    while(1)
    {
        /*** Ethernet to Serial ***/

        if (n_len == n_index)
        {
            if (inbuf != NULL)
            {
                netbuf_delete(inbuf);
                inbuf = NULL;
            }

            // net buffer is empty, try to get new data from network.
            netconn_set_recvtimeout(conn, 1);
            net_err = netconn_recv(conn, &inbuf);
            if (net_err == ERR_OK)
            {
                netbuf_data(inbuf, (void **)&n_buf, &n_len);
                n_index = 0;
            }
            else if (net_err != ERR_TIMEOUT)
            {
                printf("Socket Recv Err (%d)\r\n", n_len);
                netconn_close(conn);
                break;
            }
        }
        else
        {
            uint8_t *pBuf = UartTxBuf[buf_index];

            for(;n_index < n_len && GetTxBufLen(buf_index) < (UART_TX_BUFFER_SIZE-1); n_index++)
            {
                pBuf[UartTxHead[buf_index]] = n_buf[n_index];
                if (++UartTxHead[buf_index] >= UART_TX_BUFFER_SIZE)
                    UartTxHead[buf_index] = 0;
            }

            UART_ENABLE_INT(pmap->pserial, UART_INTEN_THREIEN_Msk);
        }

        /*** Serial to Ethernet ***/

        if ((buf_len = GetRxBufLen(buf_index)) > 0)
        {
            uint8_t *pBuf = UartRxBuf[buf_index];

            if (buf_len > sizeof(s_buf))
                buf_len = sizeof(s_buf);

            for(s_index = 0; s_index < buf_len; s_index++)
            {
                s_buf[s_index] = pBuf[UartRxTail[buf_index]];
                if (++UartRxTail[buf_index] >= UART_RX_BUFFER_SIZE)
                    UartRxTail[buf_index] = 0;
            }

            // send all data to Ethernet
            if (s_index > 0)
            {
                netconn_write(conn, s_buf, s_index, NETCONN_COPY);
                if (netconn_err(conn) != ERR_OK)
                {
                    printf("Socket Send Err (%d)\r\n", s_len);
                    netconn_close(conn);
                    break;
                }
            }
        }
    }
}

void vNetClientTask(S_PORT_CONFIG *pmap)
{
    struct netconn *conn;
    err_t err;

    printf("Thread %x in TCP client mode.\r\n", (unsigned int)pmap);

    conn = netconn_new(NETCONN_TCP);
    if (conn == NULL)
    {
        printf("TCP socket can't open (%x).\r\n", (unsigned int)pmap);
        return;
    }

    printf("Connecting server %X:%d ...\r\n", (uint32_t)(pmap->server_addr.addr), pmap->server_port);
    while(1)
    {
        err = netconn_connect(conn, &(pmap->server_addr), pmap->server_port);
        if (err == ERR_OK)
            break;
    }

    printf("\r\nConnected.");
    exchange_data(pmap, conn);
}

void vNetServerTask(S_PORT_CONFIG *pmap)
{
    struct netconn *conn, *newconn;
    err_t err;

    printf("Thread %x in TCP server mode.\r\n", (unsigned int)pmap);

    conn = netconn_new(NETCONN_TCP);
    if (conn == NULL)
    {
        printf("TCP server can't open (%x).\r\n", (unsigned int)pmap);
        return;
    }

    err = netconn_bind(conn, NULL, pmap->port);
    if (err != ERR_OK)
    {
        printf("TCP server can't bind address and port (%d)(%x).\r\n", err, (unsigned int)pmap);
        return;
    }

    netconn_listen(conn);
    while(1)
    {
        netconn_accept(conn, &newconn);
        if (newconn)
        {
            printf("Connected...\r\n");
            exchange_data(pmap, newconn);
            netconn_delete(newconn);
        }
   }
}


/*-----------------------------------------------------------*/

/*
 * Set up the hardware ready to run this demo.
 */
static void prvSetupHardware( void );
/*-----------------------------------------------------------*/

unsigned char my_mac_addr[6] = {0x00, 0x00, 0x00, 0x55, 0x66, 0x77};
struct netif netif;
static void vWebTask( void *pvParameters );

int main(void)
{
    int i;
    ip_addr_t ipaddr;
    ip_addr_t netmask;
    ip_addr_t gw;
    err_t err;
    
    /* Configure the hardware ready to run the test. */
    prvSetupHardware();

    printf("Start Serial-To-Ethernet ...\n");
    
#if (LWIP_DHCP == 1) && defined(USE_DHCP)
    
    IP4_ADDR(&ipaddr, 0,0,0,0);
    IP4_ADDR(&netmask, 0,0,0,0);
    IP4_ADDR(&gw, 0,0,0,0);
    
#elif TEST_WITH_WINPC == 1

  #if TEST_BUILD_CLIENT == 1
    IP4_ADDR(&ipaddr, 169,254,108,3);
  #else
    IP4_ADDR(&ipaddr, 169,254,108,2);
  #endif
    IP4_ADDR(&netmask, 255,255,0,0);
    IP4_ADDR(&gw, 169,254,108,1);

#else

  #if TEST_BUILD_CLIENT == 1
    IP4_ADDR(&ipaddr, 192,168,0,3);
  #else
    IP4_ADDR(&ipaddr, 192,168,0,2);
  #endif
    IP4_ADDR(&netmask, 255,255,255,0);
    IP4_ADDR(&gw, 192,168,0,1);

#endif

    tcpip_init(NULL, NULL);
    netif_add(&netif, &ipaddr, &netmask, &gw, NULL, ethernetif_init, tcpip_input);
    netif_set_default(&netif);
    netif_set_up(&netif);

#if (LWIP_DHCP == 1) && defined(USE_DHCP)
    err = dhcp_start(&netif);
    if (err != ERR_OK)
    {
        printf("DHCP error code is %d\n", err);
    }
    else
    {
        if (dhcp_supplied_address(&netif))
        {
            struct dhcp *dhcp = netif_dhcp_data(&netif);
            netif.ip_addr.addr = dhcp->offered_ip_addr.addr;
            netif.gw.addr = dhcp->offered_gw_addr.addr;
            netif.netmask.addr = dhcp->offered_sn_mask.addr;
            printf("DHCP got IP address is %s\n", ip4addr_ntoa(&(netif.ip_addr)));
        }
    }

#else
    printf("IP address is %s\n", ip4addr_ntoa(&(netif.ip_addr)));

#endif

    NVIC_SetPriority(EMAC_TX_IRQn, configLIBRARY_MAX_SYSCALL_INTERRUPT_PRIORITY + 1);
    NVIC_EnableIRQ(EMAC_TX_IRQn);
    NVIC_SetPriority(EMAC_RX_IRQn, configLIBRARY_MAX_SYSCALL_INTERRUPT_PRIORITY + 1);
    NVIC_EnableIRQ(EMAC_RX_IRQn);

//    xTaskCreate( vWebTask, "Web", TCPIP_THREAD_STACKSIZE, NULL, mainCHECK_TASK_PRIORITY, NULL );

    for(i=0; i<MAX_UART_PORTS; i++)
    {
        S_PORT_CONFIG *ps = &(port_config[i]);

        UART_Open(ps->pserial, ps->baud);
        UART_SetLineConfig(ps->pserial, ps->baud, ps->data, ps->parity, ps->stop);
        //ps->pserial->FIFO = (ps->pserial->FIFO & ~UART_FIFO_RFITL_Msk) | UART_FIFO_RFITL_4BYTES;
        //UART_SetTimeoutCnt(ps->pserial, 0x3E);

        if (port_config[i].mode == NET_SERVER_MODE)
            xTaskCreate((TaskFunction_t)vNetServerTask, "TCP SERVER", TCPIP_THREAD_STACKSIZE, &(port_config[i]), mainCHECK_TASK_PRIORITY, NULL);
        else
            xTaskCreate((TaskFunction_t)vNetClientTask, "TCP CLIENT", TCPIP_THREAD_STACKSIZE, &(port_config[i]), mainCHECK_TASK_PRIORITY, NULL);
    }

    //vStartBlockingQueueTasks( mainBLOCK_Q_PRIORITY );
    vStartPolledQueueTasks( mainQUEUE_POLL_PRIORITY );
    vStartSemaphoreTasks( mainSEM_TEST_PRIORITY );
    vStartGenericQueueTasks( tskIDLE_PRIORITY );
    vStartQueueSetTasks();

    printf("FreeRTOS is starting ...\n");

    /* Start the scheduler. */
    vTaskStartScheduler();

    /* If all is well, the scheduler will now be running, and the following line
    will never be reached.  If the following line does execute, then there was
    insufficient FreeRTOS heap memory available for the idle and/or timer tasks
    to be created.  See the memory management section on the FreeRTOS web site
    for more details. */
    for( ;; );
}
/*-----------------------------------------------------------*/

static void prvSetupHardware( void )
{
    register int i;

    /* Unlock protected registers */
    SYS_UnlockReg();

    /* Enable External XTAL (4~24 MHz) */
    CLK_EnableXtalRC(CLK_PWRCTL_HXTEN_Msk);

    /* Waiting for 12MHz clock ready */
    CLK_WaitClockReady( CLK_STATUS_HXTSTB_Msk);

    /* Set core clock as PLL_CLOCK from PLL */
    CLK_SetCoreClock(FREQ_192MHZ);

    /* Set both PCLK0 and PCLK1 as HCLK/2 */
    CLK->PCLKDIV = CLK_PCLKDIV_PCLK0DIV2 | CLK_PCLKDIV_PCLK1DIV2;

    /* Enable IP clock */
    CLK_EnableModuleClock(UART0_MODULE);
    CLK_EnableModuleClock(EMAC_MODULE);

    /* Select IP clock source */
    CLK_SetModuleClock(UART0_MODULE, CLK_CLKSEL1_UART0SEL_HXT, CLK_CLKDIV0_UART0(1));

    // Configure MDC clock rate to HCLK / (127 + 1) = 1.5 MHz if system is running at 192 MHz
    CLK_SetModuleClock(EMAC_MODULE, 0, CLK_CLKDIV3_EMAC(127));

    /* Update System Core Clock */
    /* User can use SystemCoreClockUpdate() to calculate SystemCoreClock. */
    SystemCoreClockUpdate();


    /*---------------------------------------------------------------------------------------------------------*/
    /* Init I/O Multi-function                                                                                 */
    /*---------------------------------------------------------------------------------------------------------*/
    // Configure RMII pins
    SYS->GPA_MFPL = SYS_GPA_MFPL_PA6MFP_EMAC_RMII_RXERR | SYS_GPA_MFPL_PA7MFP_EMAC_RMII_CRSDV;
    SYS->GPC_MFPL = SYS_GPC_MFPL_PC6MFP_EMAC_RMII_RXD1 | SYS_GPC_MFPL_PC7MFP_EMAC_RMII_RXD0;
    SYS->GPC_MFPH = SYS_GPC_MFPH_PC8MFP_EMAC_RMII_REFCLK;
    SYS->GPE_MFPH = SYS_GPE_MFPH_PE8MFP_EMAC_RMII_MDC |
                    SYS_GPE_MFPH_PE9MFP_EMAC_RMII_MDIO |
                    SYS_GPE_MFPH_PE10MFP_EMAC_RMII_TXD0 |
                    SYS_GPE_MFPH_PE11MFP_EMAC_RMII_TXD1 |
                    SYS_GPE_MFPH_PE12MFP_EMAC_RMII_TXEN;


    // Enable high slew rate on all RMII TX output pins
    PE->SLEWCTL = (GPIO_SLEWCTL_HIGH << GPIO_SLEWCTL_HSREN10_Pos) |
                  (GPIO_SLEWCTL_HIGH << GPIO_SLEWCTL_HSREN11_Pos) |
                  (GPIO_SLEWCTL_HIGH << GPIO_SLEWCTL_HSREN12_Pos);

    /* Set GPG multi-function pins for UART0 RXD and TXD */
    SYS->GPB_MFPH = (SYS->GPB_MFPH & ~(SYS_GPB_MFPH_PB12MFP_Msk|SYS_GPB_MFPH_PB13MFP_Msk)) |
                    (SYS_GPB_MFPH_PB12MFP_UART0_RXD | SYS_GPB_MFPH_PB13MFP_UART0_TXD);

    // Configure UART ports
    for(i=0; i<MAX_UART_PORTS; i++)
    {
        UART_T *ps = port_config[i].pserial;

        if (ps == UART0)
        {
            CLK_EnableModuleClock(UART0_MODULE);
            CLK_SetModuleClock(UART0_MODULE, CLK_CLKSEL1_UART0SEL_HXT, CLK_CLKDIV0_UART0(1));
            SYS->GPC_MFPH = (SYS->GPC_MFPH & ~(SYS_GPC_MFPH_PC11MFP_Msk|SYS_GPC_MFPH_PC12MFP_Msk)) |
                            (SYS_GPC_MFPH_PC11MFP_UART0_RXD | SYS_GPC_MFPH_PC12MFP_UART0_TXD);
            UART_ENABLE_INT(UART0, (UART_INTEN_RDAIEN_Msk | UART_INTEN_RXTOIEN_Msk | UART_INTEN_TOCNTEN_Msk /* | UART_INTEN_THREIEN_Msk */));
            NVIC_EnableIRQ(UART0_IRQn);
        }
        else if (ps == UART1)
        {
            CLK_EnableModuleClock(UART1_MODULE);
            CLK_SetModuleClock(UART1_MODULE, CLK_CLKSEL1_UART1SEL_HXT, CLK_CLKDIV0_UART1(1));
            SYS->GPH_MFPH = (SYS->GPH_MFPH & ~(SYS_GPH_MFPH_PH9MFP_Msk|SYS_GPH_MFPH_PH8MFP_Msk)) |
                            (SYS_GPH_MFPH_PH9MFP_UART1_RXD | SYS_GPH_MFPH_PH8MFP_UART1_TXD);
            UART_ENABLE_INT(UART1, (UART_INTEN_RDAIEN_Msk | UART_INTEN_RXTOIEN_Msk | UART_INTEN_TOCNTEN_Msk /* | UART_INTEN_THREIEN_Msk */));
            NVIC_EnableIRQ(UART1_IRQn);
        }
        else if (ps == UART2)
        {
            CLK_EnableModuleClock(UART2_MODULE);
            CLK_SetModuleClock(UART2_MODULE, CLK_CLKSEL3_UART2SEL_HXT, CLK_CLKDIV4_UART2(1));
            SYS->GPC_MFPL = (SYS->GPC_MFPL & ~(SYS_GPC_MFPL_PC0MFP_Msk|SYS_GPC_MFPL_PC1MFP_Msk)) |
                            (SYS_GPC_MFPL_PC0MFP_UART2_RXD | SYS_GPC_MFPL_PC1MFP_UART2_TXD);
            UART_ENABLE_INT(UART2, (UART_INTEN_RDAIEN_Msk | UART_INTEN_RXTOIEN_Msk | UART_INTEN_TOCNTEN_Msk /* | UART_INTEN_THREIEN_Msk */));
            NVIC_EnableIRQ(UART2_IRQn);
        }
        else if (ps == UART3)
        {
            CLK_EnableModuleClock(UART3_MODULE);
            CLK_SetModuleClock(UART3_MODULE, CLK_CLKSEL3_UART3SEL_HXT, CLK_CLKDIV4_UART3(1));
            SYS->GPD_MFPL = (SYS->GPD_MFPL & ~(SYS_GPD_MFPL_PD0MFP_Msk|SYS_GPD_MFPL_PD1MFP_Msk)) |
                            (SYS_GPD_MFPL_PD0MFP_UART3_RXD | SYS_GPD_MFPL_PD1MFP_UART3_TXD);
            UART_ENABLE_INT(UART3, (UART_INTEN_RDAIEN_Msk | UART_INTEN_RXTOIEN_Msk | UART_INTEN_TOCNTEN_Msk /* | UART_INTEN_THREIEN_Msk */));
            NVIC_EnableIRQ(UART3_IRQn);
        }
        else if (ps == UART4)
        {
            CLK_EnableModuleClock(UART4_MODULE);
            CLK_SetModuleClock(UART4_MODULE, CLK_CLKSEL3_UART4SEL_HXT, CLK_CLKDIV4_UART4(1));
            SYS->GPC_MFPL = (SYS->GPC_MFPL & ~(SYS_GPC_MFPL_PC4MFP_Msk|SYS_GPC_MFPL_PC5MFP_Msk)) |
                            (SYS_GPC_MFPL_PC4MFP_UART4_RXD | SYS_GPC_MFPL_PC5MFP_UART4_TXD);
            UART_ENABLE_INT(UART4, (UART_INTEN_RDAIEN_Msk | UART_INTEN_RXTOIEN_Msk | UART_INTEN_TOCNTEN_Msk /* | UART_INTEN_THREIEN_Msk */));
            NVIC_EnableIRQ(UART4_IRQn);
        }
        else //if (ps == UART5)
        {
            CLK_EnableModuleClock(UART5_MODULE);
            CLK_SetModuleClock(UART5_MODULE, CLK_CLKSEL3_UART5SEL_HXT, CLK_CLKDIV4_UART5(1));
            SYS->GPB_MFPL = (SYS->GPB_MFPL & ~(SYS_GPB_MFPL_PB4MFP_Msk|SYS_GPB_MFPL_PB5MFP_Msk)) |
                            (SYS_GPB_MFPL_PB4MFP_UART5_RXD | SYS_GPB_MFPL_PB5MFP_UART5_TXD);
            UART_ENABLE_INT(UART5, (UART_INTEN_RDAIEN_Msk | UART_INTEN_RXTOIEN_Msk | UART_INTEN_TOCNTEN_Msk /* | UART_INTEN_THREIEN_Msk */));
            NVIC_EnableIRQ(UART5_IRQn);
        }
    }

    /* Lock protected registers */
    SYS_LockReg();

    /* Init UART to 115200-8n1 for print message */
    UART_Open(UART0, 115200);
}
/*-----------------------------------------------------------*/

void vApplicationMallocFailedHook( void )
{
    /* vApplicationMallocFailedHook() will only be called if
    configUSE_MALLOC_FAILED_HOOK is set to 1 in FreeRTOSConfig.h.  It is a hook
    function that will get called if a call to pvPortMalloc() fails.
    pvPortMalloc() is called internally by the kernel whenever a task, queue,
    timer or semaphore is created.  It is also called by various parts of the
    demo application.  If heap_1.c or heap_2.c are used, then the size of the
    heap available to pvPortMalloc() is defined by configTOTAL_HEAP_SIZE in
    FreeRTOSConfig.h, and the xPortGetFreeHeapSize() API function can be used
    to query the size of free heap space that remains (although it does not
    provide information on how the remaining heap might be fragmented). */
    taskDISABLE_INTERRUPTS();
    for( ;; );
}
/*-----------------------------------------------------------*/

void vApplicationIdleHook( void )
{
    /* vApplicationIdleHook() will only be called if configUSE_IDLE_HOOK is set
    to 1 in FreeRTOSConfig.h.  It will be called on each iteration of the idle
    task.  It is essential that code added to this hook function never attempts
    to block in any way (for example, call xQueueReceive() with a block time
    specified, or call vTaskDelay()).  If the application makes use of the
    vTaskDelete() API function (as this demo application does) then it is also
    important that vApplicationIdleHook() is permitted to return to its calling
    function, because it is the responsibility of the idle task to clean up
    memory allocated by the kernel to any task that has since been deleted. */
}
/*-----------------------------------------------------------*/

void vApplicationStackOverflowHook( xTaskHandle pxTask, signed char *pcTaskName )
{
    ( void ) pcTaskName;
    ( void ) pxTask;

    /* Run time stack overflow checking is performed if
    configCHECK_FOR_STACK_OVERFLOW is defined to 1 or 2.  This hook
    function is called if a stack overflow is detected. */
    taskDISABLE_INTERRUPTS();
    for( ;; );
}
/*-----------------------------------------------------------*/

void vApplicationTickHook( void )
{
    /* This function will be called by each tick interrupt if
    configUSE_TICK_HOOK is set to 1 in FreeRTOSConfig.h.  User code can be
    added here, but the tick hook is called from an interrupt context, so
    code must not attempt to block, and only the interrupt safe FreeRTOS API
    functions can be used (those that end in FromISR()).  */

    #if ( mainCREATE_SIMPLE_BLINKY_DEMO_ONLY == 0 )
    {
        /* In this case the tick hook is used as part of the queue set test. */
        vQueueSetAccessQueueSetFromISR();
    }
    #endif /* mainCREATE_SIMPLE_BLINKY_DEMO_ONLY */
}

#if 0
static void vWebTask( void *pvParameters )
{
    ip_addr_t ipaddr;
    ip_addr_t netmask;
    ip_addr_t gw;

    IP4_ADDR(&gw, 192,168,0,1);
    IP4_ADDR(&ipaddr, 192,168,0,2);
    IP4_ADDR(&netmask, 255,255,255,0);

    tcpip_init(NULL, NULL);

    netif_add(&netif, &ipaddr, &netmask, &gw, NULL, ethernetif_init, tcpip_input);

    netif_set_default(&netif);
#if (LWIP_DHCP == 1) && defined(USE_DHCP)
    dhcp_start(&netif);
#else
    netif_set_up(&netif);
#endif
    NVIC_SetPriority(EMAC_TX_IRQn, configLIBRARY_MAX_SYSCALL_INTERRUPT_PRIORITY + 1);
    NVIC_EnableIRQ(EMAC_TX_IRQn);
    NVIC_SetPriority(EMAC_RX_IRQn, configLIBRARY_MAX_SYSCALL_INTERRUPT_PRIORITY + 1);
    NVIC_EnableIRQ(EMAC_RX_IRQn);

    http_server_netconn_init();

    vTaskSuspend( NULL );

}
#endif

/*-----------------------------------------------------------*/
